#ifndef RTL_MEMORY_H
#define RTL_MEMORY_H
#include "..\Component.h"

#ifndef MTI_SYSTEMC

template<int size = 2, int dataWidth = 2, int addressWidth = 2> class SinglePortMemory : public Component
{
public:
	sc_in< sc_lv<dataWidth> > input;
	sc_in<sc_logic> readWriteEnable;	// if true then read, if false then write
	sc_in< sc_lv<addressWidth> > address;
	sc_out< sc_lv<dataWidth> > output;
	sc_signal< sc_lv<dataWidth> > memory[size];

	SinglePortMemory(sc_module_name name) : output("output"), input("input"), readWriteEnable("rwEnable"), address("address") 
	{
		for(int i = 0; i < size; ++i)
		{
			this->memory[i].write(0);
		}
	}

	void Update()
	{
		if(this->readWriteEnable.read() == SC_LOGIC_1)
		{
			int address = this->address.read().to_uint();
			this->output.write(this->memory[address].read());
		}
		else if(this->readWriteEnable.read() == SC_LOGIC_0)
		{
			int address = this->address.read().to_uint();
			this->memory[address].write(this->input.read());
		}
	}

	void Reset()
	{
		this->output.write(0);
	}

	void Trace(sc_trace_file* tf)
	{
		sc_trace(tf, this->input, this->input.name());
		sc_trace(tf, this->output, this->output.name());
		sc_trace(tf, this->readWriteEnable, this->readWriteEnable.name());
		sc_trace(tf, this->address, this->address.name());
	}
};
template<int size = 2, int dataWidth = 2, int addressWidth = 2> class DualPortMemory : public Component
{
public:
	sc_in< sc_lv<dataWidth> > input;
	sc_out< sc_lv<dataWidth> > output;
	sc_in<sc_logic> readEnable;
	sc_in<sc_logic> writeEnable;
	sc_in< sc_lv<addressWidth> > readAddress;
	sc_in< sc_lv<addressWidth> > writeAddress;

	sc_signal< sc_lv<dataWidth> > memory[size];

	DualPortMemory(sc_module_name name)  : input("input"), readEnable("readEnable"), readAddress("readAddress"), output("output"), writeEnable("writeEnable"), writeAddress("writeAddress")
	{
		for(int i = 0; i < size; ++i)
		{
			this->memory[i].write(0);
		}
	}
	
	void Update()
	{
		if(this->readEnable.read() == SC_LOGIC_1)
		{
			int address = this->readAddress.read().to_uint();
			this->output.write(this->memory[address].read());
		}
		if(this->writeEnable.read() == SC_LOGIC_1)
		{
			int address = this->writeAddress.read().to_uint();
			this->memory[address].write(this->input.read());
		}
		if(this->writeEnable.read() == SC_LOGIC_1 && this->readEnable.read() == SC_LOGIC_1 && this->writeAddress.read() == this->readAddress.read())
		{
			this->output.write(this->input.read());
		}
	}

	void Reset()
	{
		this->output.write(0);
	}
};
template<int size = 2, int dataWidth = 2, int addressWidth = 2> class RomMemory : public Component
{
public:
	sc_in<sc_logic> readEnable;
	sc_in< sc_lv<addressWidth> > readAddress;
	sc_out< sc_lv<dataWidth> > output;

	sc_signal< sc_lv<dataWidth> > memory[size];


	RomMemory(sc_module_name name) : output("output"), readEnable("readEnable"), readAddress("readAddress")
	{
		unsigned int length = (unsigned)strlen(CONTROL_PATH) + (unsigned)strlen(name) + 5;
		char* controlFileName = new char[length];
		strcpy(controlFileName, CONTROL_PATH);
		strcat(controlFileName, name);
		strcat(controlFileName, ".mif");
		FILE* file;
#ifdef MTI_SYSTEMC
		file = fopen(controlFileName,"r");
#else
		if (fopen_s(&file, controlFileName,"r") != NULL) 
		{ 
			printf("Failed to open file containing control information.\n"); 
		}
#endif //MTI_SYSTEMC
		fseek(file, 0L, SEEK_SET);
		char str[dataWidth + 1];
		for(int i = 0; i < size; i++)
		{
			fscanf(file,"%s", str);
			sc_lv<dataWidth> temp(str);
			this->memory[i].write(temp);
		}
		fclose(file);
	}
	
	void Update()
	{
		if(this->readEnable.read() == SC_LOGIC_1)
		{
			int address = this->readAddress.read().to_uint();
			this->output.write(this->memory[address].read());
		}
	}
	void Reset()
	{
		this->output.write(0);
		// No Reset in ROM-Memory
	}
};

template<int size = 2, int dataWidth = 2, int addressWidth = 2> class RomMemory2 : public Component
{
public:
	sc_in<sc_logic> readEnable;
	sc_in< sc_lv<addressWidth> > readAddress;
	sc_out< sc_lv<dataWidth> > output;

	sc_signal< sc_lv<dataWidth> > memory[size];
	sc_signal< sc_lv<dataWidth> > output_reg;
	sc_signal< sc_lv<dataWidth> > output_reg2;

	RomMemory2(sc_module_name name) : output("output"), readEnable("readEnable"), readAddress("readAddress")
	{
		unsigned int length = (unsigned)strlen(CONTROL_PATH) + (unsigned)strlen(name) + 5;
		char* controlFileName = new char[length];
		strcpy(controlFileName, CONTROL_PATH);
		strcat(controlFileName, name);
		strcat(controlFileName, ".mif");
		FILE* file;
#ifdef MTI_SYSTEMC
		file = fopen(controlFileName,"r");
#else
		if (fopen_s(&file, controlFileName,"r") != NULL) 
		{ 
			printf("Failed to open file containing control information.\n"); 
		}
#endif //MTI_SYSTEMC
		fseek(file, 0L, SEEK_SET);
		char str[dataWidth + 1];
		for(int i = 0; i < size; i++)
		{
			fscanf(file,"%s", str);
			sc_lv<dataWidth> temp(str);
			this->memory[i].write(temp);
		}
		fclose(file);
	}
	
	void Update()
	{
		if(this->readEnable.read() == SC_LOGIC_1)
		{
			int address = this->readAddress.read().to_uint();
			this->output_reg.write(this->memory[address].read());
		}
		this->output_reg2.write(this->output_reg.read());
		this->output.write(this->output_reg2.read());
	}
	void Reset()
	{
		this->output.write(0);
		this->output_reg.write(0);
		this->output_reg2.write(0);
		// No Reset in ROM-Memory
	}
};
#else
template <int size = 2, int dataWidth = 2, int addressWidth = 2> class RomMemory : public VHDL_Component
{
public:
	sc_in<sc_logic> readEnable;
	sc_in< sc_lv<addressWidth> > readAddress;
	sc_out< sc_lv<dataWidth> > output;

    RomMemory(sc_module_name name) : VHDL_Component(name),
		output("output"), readEnable("r_enable"), readAddress("address")
    {
		this->add_parameter("size", size);
		this->add_parameter("width_data", dataWidth);
		this->add_parameter("width_address", addressWidth);

		unsigned int length = (unsigned)strlen(CONTROL_PATH) + (unsigned)strlen(name) + 12;
		char* controlFileName = new char[length];
		strcpy(controlFileName, "filename=\"");
		strcat(controlFileName, CONTROL_PATH);
		strcat(controlFileName, name);
		strcat(controlFileName, "\"");

		FILE* file;
		const char* generic_list[1];
		//generic_list[0] = strdup("filename=\"../LDPC/Monitors/rom\"");
		generic_list[0] = strdup(controlFileName);
		elaborate_foreign_module("work.rom", 1, generic_list);
    }

    ~RomMemory()
    {}
};

template <int size = 2, int dataWidth = 2, int addressWidth = 2> class DualPortMemory : public VHDL_Component
{
public:
	sc_in< sc_lv<dataWidth> > input;
	sc_out< sc_lv<dataWidth> > output;
	sc_in<sc_logic> readEnable;
	sc_in<sc_logic> writeEnable;
	sc_in< sc_lv<addressWidth> > readAddress;
	sc_in< sc_lv<addressWidth> > writeAddress;


    DualPortMemory(sc_module_name name) : VHDL_Component(name),
		output("output"), input("input"), readEnable("r_enable"), readAddress("r_address"), writeEnable("w_enable"), writeAddress("w_address")
    {
		this->add_parameter("size", size);
		this->add_parameter("width_data", dataWidth);
		this->add_parameter("width_address", addressWidth);
		elaborate_foreign_module("work.MemoryDualPort");
    }

    ~DualPortMemory()
    {}
};
template <int size = 2, int dataWidth = 2, int addressWidth = 2> class SinglePortMemory : public VHDL_Component
{
public:
	sc_in< sc_lv<dataWidth> > input;
	sc_in<sc_logic> readWriteEnable;	// if true then read, if false then write
	sc_in< sc_lv<addressWidth> > address;
	sc_out< sc_lv<dataWidth> > output;


    SinglePortMemory(sc_module_name name) : VHDL_Component(name),
		output("output"), input("input"), readWriteEnable("rw_enable"), address("address")
    {
		this->add_parameter("size", size);
		this->add_parameter("width_data", dataWidth);
		this->add_parameter("width_address", addressWidth);
		elaborate_foreign_module("work.memorysingleport");
    }

    ~SinglePortMemory()
    {}
};
#endif //MTI_SYSTEMC

#endif //RTL_MEMORY_H
