#ifndef VARIABLESHIFTREGISTER_H
#define VARIABLESHIFTREGISTER_H
#include "..\Component.h"

#ifndef MTI_SYSTEMC
template<class T>class ShiftRegister : public Component
{
public:
	sc_signal<T>* shift_register;
	sc_in<T> input;
	sc_out<T> output;
	int size;

	ShiftRegister(sc_module_name name, int delay, int width)  :
		input("input"), output("output")
	{
		this->size = delay - 1;
		this->shift_register = new sc_signal<T>[this->size];
	}

	void Update()
	{
		if(this->size == 0)
		{
			this->output.write(this->input.read());
		}
		else
		{
			this->output.write(this->shift_register[this->size - 1]);
			for(int i = 1; i < this->size; i++)
			{
				this->shift_register[i].write(this->shift_register[i-1].read());
			}
			this->shift_register[0].write(this->input.read());
		}
	}

	void Reset()
	{
		T temp = (T) 0;
		for(int i = 0; i < size; i++)
		{
			this->shift_register[i].write(temp);
		}
		this->output.write(temp);
	}

	void TraceSignals(sc_trace_file *tf)
	{
		sc_trace(tf, this->input, "input");
		sc_trace(tf, this->output, "output");
		for(int i = 0; i < this->size; i++)
		{
			sc_trace(tf, this->shift_register[i], PosIntToString("Register_",i));
		}
	}
};

// this shift register has fixed output at the last address and variable input 
// wich is calculated through active size and offset
template<class T>class VariableShiftRegister : public Component
{
public:
	sc_in<T> input;
	sc_out<T> output;
	sc_in<sc_lv<DEGREE_WIDTH> > active_size;

	sc_signal<T>* shift_register;
	int size;
	int offset;							// additional difference to the active size

	VariableShiftRegister(sc_module_name name, int size, int offset, int width) : 
		input("input"), output("output"), active_size("active_size")
	{
		this->offset = offset;
		this->size = size + offset - 1;
		this->shift_register = new sc_signal<T>[this->size];
	}

	void Update()
	{
		for(int i = 1; i < this->size; i++)
		{
			this->shift_register[i].write(this->shift_register[i-1].read());
		}

		int delay = active_size.read().to_uint() + this->offset;
		if(delay > 0) // delay == 0 means delay of one 
		{
			this->output.write(this->shift_register[this->size - 1]);
			this->shift_register[this->size - delay].write(this->input.read());
		}
		else
		{
			this->output.write(this->input.read());
		}
	}

	void Reset()
	{
		T temp = (T) 0;
		for(int i = 0; i < size; i++)
		{
			this->shift_register[i].write(temp);
		}
		this->output.write(temp);
	}

	void TraceSignals(sc_trace_file *tf)
	{
		sc_trace(tf, this->input, "input");
		sc_trace(tf, this->output, "output");
		sc_trace(tf, this->active_size, "activeSize");
		for(int i = 0; i < this->size; i++)
		{
			sc_trace(tf, this->shift_register[i], PosIntToString("Register_",i));
		}
	}
};
// VariableShiftRegister with Read Signal to get the content.
template<class T>class VariableShiftRegisterCN : public Component
{
public:
	sc_in<T> input;
	sc_out<T> output;
	sc_in<sc_lv<DEGREE_WIDTH> > active_size;
	
	sc_signal<T>* shift_register;
	int size;
	int offset;							// additional difference to the active size

	VariableShiftRegisterCN(sc_module_name name, int size, int offset, int width) : 
		input("input"), output("output"), active_size("active_size")
	{
		this->offset = offset;
		this->size = size + offset - 1;
		this->shift_register = new sc_signal<T>[this->size];
	}

	void Update()
	{
		for(int i = 1; i < this->size; i++)
		{
			this->shift_register[i].write(this->shift_register[i-1].read());
		}

		int delay = active_size.read().to_uint() + this->offset;
		if(delay > 0) // delay == 0 means delay of one 
		{
			this->output.write(this->shift_register[this->size - 1]);
			this->shift_register[this->size - delay].write(this->input.read());
		}
		else
		{
			this->output.write(this->input.read());
		}
	}

	void Reset()
	{
		T temp = (T) 0;
		for(int i = 0; i < size; i++)
		{
			this->shift_register[i].write(temp);
		}
		this->output.write(temp);
	}

	void TraceSignals(sc_trace_file *tf)
	{
		sc_trace(tf, this->input, "input");
		sc_trace(tf, this->output, "output");
		sc_trace(tf, this->active_size, "activeSize");
		for(int i = 0; i < this->size; i++)
		{
			sc_trace(tf, this->shift_register[i], PosIntToString("Register_",i));
		}
	}
};

#else

template <class T = int> class ShiftRegister : public VHDL_Component
{
public:
    sc_in<T > input;
    sc_out<T > output;

    ShiftRegister(sc_module_name nm, int size, int width)
     : VHDL_Component(nm),
       input("input"),
       output("output")
    {
		this->add_parameter("size", size);
		this->add_parameter("width", width);
		elaborate_foreign_module("work.shiftregister");
    }
};
template <class T = int> class VariableShiftRegister : public VHDL_Component
{
public:
	sc_in<sc_lv<3> > active_size;
    sc_in<T > input;
    sc_out<T > output;


    VariableShiftRegister(sc_module_name nm, int size, int offset, int width)
     : VHDL_Component(nm),
	   active_size("active_size"),
       input("input"),
       output("output")
    {
		this->add_parameter("size", size);
		this->add_parameter("width", width);
		this->add_parameter("offset", offset);
		elaborate_foreign_module("work.variableshiftregister");
    }
};

template <class T = int> class VariableShiftRegisterCN : public VHDL_Component
{
public:
	sc_in<sc_lv<3> > active_size;
	sc_in<sc_logic> read_enable;
	sc_out<sc_lv<MAX_DEGREE> > content;
    sc_in<T > input;
    sc_out<T > output;


    VariableShiftRegisterCN(sc_module_name nm, int size, int offset, int width)
     : VHDL_Component(nm),
	   active_size("active_size"),
       input("input"),
       output("output"),
	   content("content"),
	   read_enable("read_enable")
    {
		this->add_parameter("size", size);
		elaborate_foreign_module("work.VariableShiftRegisterCN");
    }
};

#endif //MTI_SYSTEMC*/

#endif //VARIABLESHIFTREGISTER_H
