-- Filename    	: FiniteStateMachine.vhd
-- Author:	   	: Andreas Torno
-- Description 	: This component implements finite state machine

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
library work;
USE work.widths.all;

ENTITY FiniteStateMachine IS
	GENERIC (
		WIDTH_ITERATION				: integer := 9
	);
    PORT (
		-- Global control ports
		clock						: IN  std_logic;
		enable						: IN  std_logic;
		reset						: IN  std_logic;
		
		-- Input interface
		in_newValuesPresent			: IN  std_logic;
		out_readyForNextCodeword	: OUT std_logic;
		
		-- Output interface
		out_decodingValid			: OUT std_logic;
		out_iterations				: OUT std_logic_vector(WIDTH_ITERATION - 1 downto 0);
		
		-- Control Signals
		in_decodingValid			: IN  std_logic;
		in_currentIteration			: IN  std_logic_vector(WIDTH_ITERATION - 1 downto 0);
		out_enableDecodeSchedulig	: OUT std_logic;
		out_resetDecodeScheduling	: OUT std_logic;
		in_lastIterationDone		: IN  std_logic;
		in_initAndOutputDone		: IN  std_logic;
		out_initState				: OUT std_logic;
		out_outputState				: OUT std_logic
    );
END FiniteStateMachine;

ARCHITECTURE RTL OF FiniteStateMachine IS
	TYPE state IS (IDLE, INIT, DECODE, WRITEBACK, INIT_WRITEBACK);
	signal next_state, current_state	: state;
	
	signal readyForNextCodeword			: std_logic;
	signal enableDecodeSchedulig		: std_logic;
	signal resetDecodeScheduling		: std_logic;
	signal initState					: std_logic;
	signal outputState				: std_logic;
BEGIN

	registers : process(clock)
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				current_state 				<= IDLE;
				out_resetDecodeScheduling	<= '1';
			else
				current_state 				<= next_state;
				out_readyForNextCodeword	<= readyForNextCodeword;
				out_initState				<= initState;
				out_enableDecodeSchedulig	<= enableDecodeSchedulig;
				out_resetDecodeScheduling	<= resetDecodeScheduling;
				out_outputState				<= outputState;
				if (in_lastIterationDone or in_decodingValid) = '1' then
					out_decodingValid		<= in_decodingValid;
					out_iterations			<= in_currentIteration;
				end if;
			end if;
		end if;
	END process;
	
	mealey : process (current_state, in_newValuesPresent, in_decodingValid, in_lastIterationDone, in_initAndOutputDone)
	BEGIN
		readyForNextCodeword			<= '0';
		initState						<= '0';
		enableDecodeSchedulig			<= '0';
		resetDecodeScheduling			<= '0';
		next_state						<= current_state;
		outputState				<= '0';
		
		case current_state is
		when IDLE =>
			if in_newValuesPresent = '1' then
				readyForNextCodeword	<= '1';
				next_state				<= INIT;
			end if;
		when INIT =>
			initState					<= '1';
			if in_initAndOutputDone = '1' then
				next_state 				<= DECODE;
			end if;
		when DECODE =>
			if (in_lastIterationDone or in_decodingValid) = '1' then
				resetDecodeScheduling	<= '1';
				if in_newValuesPresent = '1' then 
					next_state			<= INIT_WRITEBACK;
					readyForNextCodeword<= '1';
				else
					next_state			<= WRITEBACK;
				end if;
			else
				enableDecodeSchedulig	<= '1';
			end if;
		when WRITEBACK =>
			outputState 				<= '1';
			if in_initAndOutputDone = '1' then
				next_state 				<= IDLE;
			end if;
		when INIT_WRITEBACK =>
			initState					<= '1';
			outputState					<= '1';
			if in_initAndOutputDone = '1' then
				next_state 				<= DECODE;
			end if;
		end case;
	END process;

END RTL;