-- Filename    : VariableShiftRegister.vhd
-- Author:	   : Andreas Torno
-- Description : Generic variable shift register with variable length dependent on
--			   : the signal active_size wich reflects the check node degree

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
library work;
USE work.widths.all;

ENTITY VariableShiftRegister IS
	GENERIC (
		size 		: integer := 1;
		width 		: integer := 32;
		offset		: integer := 1
	);
	
    PORT (
		-- Control ports
		clock		: IN std_logic;
		enable		: IN std_logic;
		reset		: IN std_logic;
		active_size : IN std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		
		-- Data ports
        input     	: IN std_logic_vector(width-1 downto 0);
        output    	: OUT std_logic_vector(width-1 downto 0)
    );
END VariableShiftRegister;

ARCHITECTURE RTL OF VariableShiftRegister IS
	type var_shift_type is array(size - 1 downto 0) 		of std_logic_vector(width-1 downto 0);
	type shift_type 	is array(offset - 1 downto 0) 	of std_logic_vector(width-1 downto 0);
	signal var_registers	: var_shift_type;
	signal registers 		: shift_type;
BEGIN
	VariableShiftRegister: process(clock)
    begin
	if clock'event and clock='1' then
		if reset = '1' then
			for i in 0 to size - 1 LOOP
				var_registers(i) <= (others => '0');
			end LOOP;
	    elsif enable = '1' then
			for i in 1 to size - 1 LOOP
				var_registers(i - 1) <= var_registers(i);
			end LOOP;
			var_registers(to_integer(unsigned(active_size))) <= input;
	    end if;
	end if;
    end process;
	
	WithShiftRegister: if offset > 0 generate
		process(clock)
		begin
		if clock'event and clock='1' then
			if reset = '1' then
				for i in 0 to offset - 1 LOOP
					registers(i) <= (others => '0');
				end LOOP;
			elsif enable = '1' then
				registers(offset - 1)	<= var_registers(0);
				for i in 1 to offset - 1 LOOP
					registers(i - 1) <= registers(i);
				end LOOP;
			end if;
		end if;
		end process;
		output 				<= registers(0);
	end generate;
	
	WithoutShiftRegister: if offset = 0 generate
		output				<= var_registers(0);
	end generate;
END RTL;
