-- Filename     : Decoder.vhd
-- Author:	    : Andreas Torno
-- Description  : This is the FPGA implementation of the OPMP-LDPC-Decoder
library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
library work;
USE work.delays.all;
USE work.widths.all;
USE work.types.all;

ENTITY Decoder IS
    PORT (
		-- Global control ports
		clock								: IN  std_logic;
		enable								: IN  std_logic;
		reset								: IN  std_logic;
		
		-- Input interface
		in_received_value					: IN  std_logic_vector(WIDTH_SD - 1 downto 0);
		in_new_values_present				: IN  std_logic;
		out_ready_for_next_received_value	: OUT std_logic;
		
		-- Output inteface
		out_iterations						: OUT std_logic_vector(WIDTH_ITERATION - 1 downto 0);
		out_decoded_value					: OUT std_logic;
		out_decoded_value_present			: OUT std_logic;
		out_decoding_valid					: OUT std_logic	
    );
END Decoder;

ARCHITECTURE RTL OF Decoder IS
	TYPE SD_VALUES IS ARRAY (0 to PARALLELISM - 1) OF std_logic_vector(WIDTH_SD - 1 DOWNTO 0);
	signal oddIteration	: std_logic;
	
	COMPONENT ShiftRegister
	GENERIC (
		size 		: integer;
		width 		: integer
	);
    PORT (
		-- Control ports
		clock		: IN std_logic;
		enable		: IN std_logic;
		reset		: IN std_logic;
		
		-- Data ports
        input     	: IN std_logic_vector(width-1 downto 0);
        output    	: OUT std_logic_vector(width-1 downto 0)
    );
	END COMPONENT;
	
	COMPONENT FiniteStateMachine
	GENERIC (
		WIDTH_ITERATION						: integer := 9
	);
    PORT (
		-- Global control ports
		clock								: IN  std_logic;
		enable								: IN  std_logic;
		reset								: IN  std_logic;
		-- Input interface
		in_newValuesPresent					: IN  std_logic;
		out_readyForNextCodeword			: OUT std_logic;
		-- Output interface
		out_decodingValid					: OUT std_logic;
		out_iterations						: OUT std_logic_vector(WIDTH_ITERATION - 1 downto 0);
		-- Control Signals
		in_decodingValid					: IN  std_logic;
		in_currentIteration					: IN  std_logic_vector(WIDTH_ITERATION - 1 downto 0);
		out_enableDecodeSchedulig			: OUT std_logic;
		out_resetDecodeScheduling			: OUT std_logic;
		in_lastIterationDone				: IN  std_logic;
		in_initAndOutputDone				: IN  std_logic;
		out_initState						: OUT std_logic;
		out_outputState						: OUT std_logic
    );
	END COMPONENT;
	signal	fsm_enable_decode_scheduling	: std_logic;
	signal	fsm_reset_ds					: std_logic;
	signal	initState						: std_logic;
	signal	writeBackState					: std_logic;
	
	
	COMPONENT InitAndOutput
    PORT (
		-- Global control ports
		clock								: IN  std_logic;
		enable								: IN  std_logic;
		reset								: IN  std_logic;
		-- Specific control ports
		in_initState						: IN  std_logic;
		in_outputState						: IN  std_logic;
		out_loadAndOutputDone				: OUT std_logic;
		out_init							: OUT std_logic_vector(PARALLELISM - 1 downto 0);
		out_output							: OUT std_logic_vector(PARALLELISM - 1 downto 0);
		out_address							: OUT std_logic_vector(ROM_SN_WIDTH - 3 downto 0)
    );
	END COMPONENT;
	signal	loadAndwriteBackDone			: std_logic;
	signal	sn_init							: std_logic_vector(PARALLELISM - 1 downto 0);
	signal	sn_output						: std_logic_vector(PARALLELISM - 1 downto 0);
	signal	sn_address						: std_logic_vector(ROM_SN_WIDTH - 3 downto 0);
	signal	received_value					: std_logic_vector(WIDTH_SD - 1 DOWNTO 0);
	signal	write_to_output_enable			: std_logic_vector(PARALLELISM - 1 downto 0);
	
	
	COMPONENT DecodeScheduling
	GENERIC (
		widthIteration						: integer := WIDTH_ITERATION;
		maxIteration						: integer := MAX_ITERATION
	);
    PORT (
		-- Global control ports
		clock								: IN std_logic;
		enable								: IN std_logic;
		reset								: IN std_logic;
		-- Specific control ports
		out_rom_index						: OUT std_logic_vector(ROM_ADDRESS_WIDTH - 1 downto 0);
		out_rom_enable						: OUT std_logic;
		out_first_value						: OUT std_logic;
		out_cn_index						: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_cn_degree						: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_iteration						: OUT std_logic_vector(widthIteration - 1 downto 0);
		out_lastIterationReached			: OUT std_logic
    );
	END COMPONENT;
	signal	ds_reset			: std_logic;
	signal	ds_cn_first_value_out			: std_logic;
	signal	ds_MEM_enable					: std_logic;
	signal	lastIterationReached			: std_logic;
	signal	ds_control_MEM_address			: std_logic_vector(ROM_ADDRESS_WIDTH - 1 downto 0);
	signal	ds_cn_degree					: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal	ds_cn_index						: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal	iteration						: std_logic_vector(WIDTH_ITERATION - 1 downto 0);
	signal	variableDelay					: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal	firstIteration					: std_logic;
		
		
	COMPONENT ControlSNAndPi
    PORT (
		-- Global control ports
		clock								: IN  std_logic;
		enable								: IN  std_logic;
		reset								: IN  std_logic;
		-- Input control ports
		in_enableRom						: IN  std_logic;
		in_oddIteration						: IN  std_logic;
		in_variableDelay					: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_addressRom						: IN  std_logic_vector(ROM_ADDRESS_WIDTH - 1 downto 0);
		-- Output control ports
		out_newSD_readEnable				: OUT std_logic;
		out_oldSD_readEnable				: OUT std_logic;
		out_oldSD_address_vector			: OUT std_logic_vector((ROM_SN_WIDTH - 2) * PARALLELISM - 1 downto 0);
		out_newSD_address_vector			: OUT std_logic_vector(ROM_SN_WIDTH * PARALLELISM - 1 downto 0);
		out_permutationIndices_vector		: OUT std_logic_vector(4 * ROM_PERMUTATION_WIDTH - 1 downto 0)
    );
	END COMPONENT;
	signal	newSD_readEnable				: std_logic;
	signal	oldSD_readEnable				: std_logic;
	signal	oldSD_address_vector			: std_logic_vector((ROM_SN_WIDTH - 2) * PARALLELISM - 1 downto 0);
	signal	romSN_output					: std_logic_vector(ROM_SN_WIDTH * PARALLELISM - 1 downto 0);
	signal	permutation_networks_indices_vector						: std_logic_vector(4 * ROM_PERMUTATION_WIDTH - 1 downto 0);
	
	
	COMPONENT ControlCN
    PORT (
		-- Global control ports
		clock								: IN  std_logic;
		enable								: IN  std_logic;
		reset								: IN  std_logic;
		-- Specific control ports
		in_firstValue						: IN  std_logic;
		in_oddIteration						: IN  std_logic;
		in_variableDelay					: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_degreeIndex						: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_valid							: IN  std_logic_vector(PARALLELISM - 1 downto 0);
		-- CN control ports
		out_firstValue						: OUT std_logic;
		out_lastValue						: OUT std_logic;
		out_degreeIndexIn					: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_degreeIndexOut					: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_variableDelay					: OUT std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_valid							: OUT std_logic;
		-- Extrinsic memory control ports
		out_memoryEnable					: OUT std_logic;
		out_memoryRead						: OUT std_logic;
		out_memoryAddress					: OUT std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0)
    );
	END COMPONENT;
	signal	cn_first_value					: std_logic;
	signal	cn_last_value					: std_logic;
	signal	cn_index_in						: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal	cn_index_out					: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal	cn_variableDelay				: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal	decoding_valid					: std_logic;
	signal	cn_MEM_enable					: std_logic;
	signal	cn_mem_read						: std_logic;
	signal	cn_mem_address					: std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0);

	
	COMPONENT CheckNode
	PORT (
		-- Global control ports
		enable, clock, reset				: IN  std_logic;
		-- Message ports
		in_input 							: IN  std_logic_vector(WIDTH_SD - 1 downto 0);
		out_output 							: OUT std_logic_vector(WIDTH_SD - 1 downto 0);	
		-- Decoding control ports
		in_indexIn							: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_indexOut							: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_firstValue						: IN  std_logic;
		in_lastValue						: IN  std_logic;
		in_firstIteration					: IN  std_logic;
		in_variableDelay					: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_valid							: OUT std_logic;	
		-- Memory control ports
		in_enableMEM						: IN  std_logic;
		in_readMEM							: IN  std_logic;
		in_addressMEM						: IN  std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0)
	);
	END COMPONENT;
	signal	cn_soft_decision				: SD_VALUES;
	signal	cn_extrinsic_message			: SD_VALUES;
	signal	cn_valid						: std_logic_vector(0 to PARALLELISM - 1);


	COMPONENT SymbolNode
	PORT (
		-- Global control ports
		enable, clock, reset				: IN  std_logic;
		-- Decoding control ports
		in_initialization					: IN  std_logic;
		in_saveValues						: IN  std_logic;
		in_oddIteration						: IN  std_logic;
		in_firstExtrinsic					: IN  std_logic;
		in_lastExtrinsic					: IN  std_logic;
		-- Memory control ports
		in_readEnable						: IN  std_logic;
		in_writeEnable						: IN  std_logic;
		in_oldSD_address					: IN  std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
		in_newSD_address					: IN  std_logic_vector(RAM_SN_ADDRESS_WIDTH - 1 downto 0);
		-- Soft Decisions ports
		in_input 							: IN  std_logic_vector(WIDTH_SD - 1 downto 0);
		out_output 							: OUT std_logic_vector(WIDTH_SD - 1 downto 0);
		in_SD_read							: IN  std_logic_vector(WIDTH_SD - 1 downto 0);
		out_SD_read							: OUT std_logic_vector(WIDTH_SD - 1 downto 0);
		in_SD_write							: IN  std_logic_vector(WIDTH_SD - 1 downto 0);
		out_SD_write						: OUT std_logic_vector(WIDTH_SD - 1 downto 0)
	);
	END COMPONENT;
	signal	sn_soft_decision					: SD_VALUES;
	signal	sn_soft_decision_delayed			: SD_VALUES;
	TYPE	SN_ADDRESSES IS ARRAY (0 to PARALLELISM - 1) OF std_logic_vector((ROM_SN_WIDTH - 2) - 1 DOWNTO 0);
	signal	oldSD_address						: SN_ADDRESSES;
	signal	sn_input							: SD_VALUES;
	
	COMPONENT PermutationNetwork
    PORT (
		clock								: IN  std_logic;
		enable								: IN  std_logic;
		reset								: IN  std_logic;
        in_input							: IN  std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
		in_index							: IN  std_logic_vector(ROM_PERMUTATION_WIDTH - 1 downto 0);
        out_output  						: OUT std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0)
    );
	END COMPONENT;
	signal	pi0_input						: std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
	signal	pi0_output						: std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
	signal	pi1_input						: std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
	signal	pi1_output						: std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
	signal	pi2_input						: std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
	signal	pi2_output						: std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
	signal	pi3_input						: std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);
	signal	pi3_output						: std_logic_vector(PARALLELISM * WIDTH_SD - 1 DOWNTO 0);

BEGIN
	FSM : FiniteStateMachine
	GENERIC MAP(
		WIDTH_ITERATION						=> WIDTH_ITERATION
	)
    PORT MAP(
		-- Global control ports
		clock								=> clock,
		enable								=> enable,
		reset								=> reset,
		-- Input interface
		in_newValuesPresent					=> in_new_values_present,
		out_readyForNextCodeword			=> out_ready_for_next_received_value,
		-- Output interface
		out_decodingValid					=> out_decoding_valid,
		out_iterations						=> out_iterations,
		-- Control Signals
		in_decodingValid					=> decoding_valid,
		in_currentIteration					=> iteration,
		out_enableDecodeSchedulig			=> fsm_enable_decode_scheduling,
		out_resetDecodeScheduling			=> fsm_reset_ds,
		in_lastIterationDone				=> lastIterationReached,
		in_initAndOutputDone				=> loadAndwriteBackDone,
		out_initState						=> initState,
		out_outputState						=> writeBackState
    );

	init_and_output : InitAndOutput
    PORT MAP(
		-- Global control ports
		clock								=> clock,
		enable								=> enable,
		reset								=> reset,
		-- Specific control ports
		in_initState						=> initState,
		in_outputState						=> writeBackState,
		out_loadAndOutputDone				=> loadAndwriteBackDone,
		out_init							=> sn_init,
		out_output							=> sn_output,
		out_address							=> sn_address
    );
	
	outputOfDecodedValues : process(clock)
	variable temp	: std_logic_vector(WIDTH_SD - 1 downto 0);
	begin
	if clock'event and clock='1' then
		if reset = '1' then
			out_decoded_value_present	<= '0';
			out_decoded_value			<= '0';
		elsif enable = '1' then
			out_decoded_value_present	<= '0';
			out_decoded_value			<= '0';
			FOR i IN 0 TO PARALLELISM - 1 LOOP
				if write_to_output_enable(i) = '1' then
					out_decoded_value_present	<= '1';
					-- temp := sn_soft_decision(i);
				-- end if;
				-- out_decoded_value <= temp(WIDTH_SD - 1);
					out_decoded_value <= sn_soft_decision(i)(WIDTH_SD - 1);
				end if;
			END LOOP;
		end if;
	end if;
	end process;
	
	ds_reset <= fsm_reset_ds or reset;
	ds : DecodeScheduling
    PORT MAP(
		-- Global control ports
		clock								=> clock,
		enable								=> fsm_enable_decode_scheduling,
		reset								=> ds_reset,
		-- Specific control ports
		out_rom_index						=> ds_control_MEM_address,
		out_rom_enable						=> ds_MEM_enable,
		out_first_value						=> ds_cn_first_value_out,
		out_cn_index						=> ds_cn_index,
		out_cn_degree						=> ds_cn_degree,
		out_iteration						=> iteration,
		out_lastIterationReached			=> lastIterationReached
    );
	oddIteration			<= iteration(0);
	
	setting_firstIteration : process(clock)
	begin
	if clock'event and clock='1' then
		if reset = '1' then
			firstIteration	<= '0';
	    elsif enable = '1' and unsigned(iteration) = 0 then
			firstIteration	<= '1';
		else
			firstIteration	<= '0';
	    end if;
	end if;
	end process;
	
	setting_variableDelay : process(clock)
	begin
	if clock'event and clock='1' then
		if reset = '1' then
			variableDelay	<= (OTHERS => '1');
	    elsif enable = '1' and ds_cn_first_value_out = '1' then
			variableDelay	<= ds_cn_degree;
	    end if;
	end if;
	end process;
	
	ROM_Access : ControlSNAndPi
    PORT MAP(
		-- Global control ports
		clock								=> clock,
		enable								=> enable,
		reset								=> reset,
		-- Input control ports
		in_enableRom						=> ds_MEM_enable,
		in_oddIteration						=> iteration(0),
		in_variableDelay					=> variableDelay,
		in_addressRom						=> ds_control_MEM_address,
		-- Output control ports
		out_newSD_readEnable				=> newSD_readEnable,
		out_oldSD_readEnable				=> oldSD_readEnable,
		out_oldSD_address_vector			=> oldSD_address_vector,
		out_newSD_address_vector			=> romSN_output,
		out_permutationIndices_vector		=> permutation_networks_indices_vector
    );
	
	SymbolNodes : FOR i IN 0 to PARALLELISM - 1 GENERATE
	BEGIN
		SN : SymbolNode
		PORT MAP(
			-- Global control ports
			clock								=> clock,
			enable								=> enable,
			reset								=> reset,
			-- Decoding control ports
			in_initialization					=> sn_init(i),
			in_saveValues						=> sn_output(i),
			in_oddIteration						=> iteration(0),
			in_firstExtrinsic					=> romSN_output((PARALLELISM - i - 1)*ROM_SN_WIDTH + 1),
			in_lastExtrinsic					=> romSN_output((PARALLELISM - i - 1)*ROM_SN_WIDTH),
			-- Memory control ports
			in_readEnable						=> oldSD_readEnable,
			in_writeEnable						=> newSD_readEnable,
			in_oldSD_address					=> oldSD_address(i),
			in_newSD_address					=> romSN_output((PARALLELISM - i)*ROM_SN_WIDTH - 1 downto (PARALLELISM - i - 1)*ROM_SN_WIDTH + 2),
			-- Soft Decisions ports
			in_input 							=> sn_input(i),
			out_output 							=> sn_soft_decision(i),
			in_SD_read							=> pi2_output((PARALLELISM - i)*WIDTH_SD - 1 downto (PARALLELISM - i - 1)*WIDTH_SD),
			out_SD_read							=> pi2_input((PARALLELISM - i)*WIDTH_SD - 1 downto (PARALLELISM - i - 1)*WIDTH_SD),
			in_SD_write							=> pi3_output((PARALLELISM - i)*WIDTH_SD - 1 downto (PARALLELISM - i - 1)*WIDTH_SD),
			out_SD_write						=> pi3_input((PARALLELISM - i)*WIDTH_SD - 1 downto (PARALLELISM - i - 1)*WIDTH_SD)
		);
		sn_input(i)				<= received_value 	when sn_init(i) = '1'						else pi1_output((PARALLELISM - i)*WIDTH_SD - 1 downto (PARALLELISM - i - 1)*WIDTH_SD);
		pi0_input((PARALLELISM - i)*WIDTH_SD - 1 downto (PARALLELISM - i - 1)*WIDTH_SD)			<= pi2_output((PARALLELISM - i)*WIDTH_SD - 1 downto (PARALLELISM - i - 1)*WIDTH_SD)	when iteration(0) = '1'						else sn_soft_decision_delayed(i);
		oldSD_address(i)		<= sn_address		when sn_init(i) = '1' or sn_output(i) = '1'	else oldSD_address_vector((PARALLELISM - i) * (ROM_SN_WIDTH - 2) - 1 downto (PARALLELISM - i - 1) * (ROM_SN_WIDTH - 2));
		
		delay_SN_output : ShiftRegister 
		GENERIC MAP(DURATION_PERMUTATION, WIDTH_SD) 
		PORT MAP(clock, enable, reset, sn_soft_decision(i), sn_soft_decision_delayed(i));
		
		delay_SN_writeToOutput : ShiftRegister
		GENERIC MAP(
			size 		=> 1 + DURATION_SN_OUTPUT,
			width 		=> 1
		)
		PORT MAP(
			clock		=> clock,
			enable		=> enable,
			reset		=> reset,
			input(0)	=> sn_output(i),
			output(0)	=> write_to_output_enable(i)
		);
	END GENERATE SymbolNodes;
	
	delay_input : ShiftRegister 
	GENERIC MAP(2 + DURATION_ROM_ACCESS, WIDTH_SD) 
	PORT MAP(clock, enable, reset, in_received_value, received_value);
	
	control_cn : ControlCN
    PORT MAP(
		-- Global control ports
		clock								=> clock,
		enable								=> enable,
		reset								=> reset,
		-- Specific control ports
		in_firstValue						=> ds_cn_first_value_out,
		in_oddIteration						=> iteration(0), 
		in_variableDelay					=> variableDelay,
		in_degreeIndex						=> ds_cn_index,
		in_valid							=> cn_valid,
		-- CN control ports
		out_firstValue						=> cn_first_value,
		out_lastValue						=> cn_last_value,
		out_degreeIndexIn					=> cn_index_in,
		out_degreeIndexOut					=> cn_index_out,
		out_variableDelay					=> cn_variableDelay,
		out_valid							=> decoding_valid,
		-- Extrinsic memory control ports
		out_memoryEnable					=> cn_MEM_enable,
		out_memoryRead						=> cn_mem_read,
		out_memoryAddress					=> cn_mem_address
    );

	CheckNodes : FOR i IN 0 to PARALLELISM - 1 GENERATE
	BEGIN
		CN : CheckNode
		PORT MAP(
			-- Global control ports
			clock								=> clock,
			enable								=> enable,
			reset								=> reset,
			-- Message ports
			in_input 							=> pi0_output((PARALLELISM - i)*WIDTH_SD - 1 downto (PARALLELISM - i - 1)*WIDTH_SD),
			out_output 							=> pi1_input((PARALLELISM - i)*WIDTH_SD - 1 downto (PARALLELISM - i - 1)*WIDTH_SD),
			-- Decoding control ports
			in_indexIn							=> cn_index_in,
			in_indexOut							=> cn_index_out,
			in_firstValue						=> cn_first_value,
			in_lastValue						=> cn_last_value,
			in_firstIteration					=> firstIteration,
			in_variableDelay					=> cn_variableDelay,
			out_valid							=> cn_valid(i),
			-- Memory control ports
			in_enableMEM						=> cn_MEM_enable,
			in_readMEM							=> cn_mem_read,
			in_addressMEM						=> cn_mem_address
		);
	END GENERATE CheckNodes;
	
	PN0 : PermutationNetwork PORT MAP(clock, enable, reset, pi0_input, permutation_networks_indices_vector(4 * ROM_PERMUTATION_WIDTH - 1 downto 3 * ROM_PERMUTATION_WIDTH), pi0_output);
	PN1 : PermutationNetwork PORT MAP(clock, enable, reset, pi1_input, permutation_networks_indices_vector(3 * ROM_PERMUTATION_WIDTH - 1 downto 2 * ROM_PERMUTATION_WIDTH), pi1_output);
	PN2 : PermutationNetwork PORT MAP(clock, enable, reset, pi2_input, permutation_networks_indices_vector(2 * ROM_PERMUTATION_WIDTH - 1 downto 1 * ROM_PERMUTATION_WIDTH), pi2_output);
	PN3 : PermutationNetwork PORT MAP(clock, enable, reset, pi3_input, permutation_networks_indices_vector(1 * ROM_PERMUTATION_WIDTH - 1 downto 0 * ROM_PERMUTATION_WIDTH), pi3_output);
	
END RTL;