--
-- Implementation Of CheckNode
-- by Andreas Torno
--   

LIBRARY ieee;
USE ieee.std_logic_1164.all;
use ieee.numeric_std.all;

LIBRARY work;
USE work.types.all;
USE work.widths.all;
USE work.delays.all;

ENTITY CheckNode IS
	PORT (
		-- Global control ports
		enable, clock, reset: IN  std_logic;
		
		-- Message ports
		in_input 			: IN  std_logic_vector(WIDTH_SD - 1 downto 0);
		out_output 			: OUT std_logic_vector(WIDTH_SD - 1 downto 0);
		
		-- Decoding control ports
		in_indexIn			: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_indexOut			: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		in_firstValue		: IN  std_logic;
		in_lastValue		: IN  std_logic;
		in_firstIteration	: IN  std_logic;
		in_variableDelay	: IN  std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		out_valid			: OUT std_logic;
		
		-- Memory control ports
		in_enableMEM		: IN  std_logic;
		in_readMEM			: IN  std_logic;
		in_addressMEM		: IN  std_logic_vector(RAM_CN_ADDRESS_WIDTH - 1 downto 0)
	);
END;

ARCHITECTURE Beh OF CheckNode IS
	constant WIDTH_MAGNITUDE: integer := WIDTH_SD - 1;
	constant CN_RAM_WIDTH 	: integer := 2*WIDTH_MAGNITUDE + DEGREE_WIDTH + MAX_DEGREE + 1;
	
	COMPONENT MemorySinglePort
		GENERIC (
			size 			: integer := CN_PER_FU;
			width_data 		: integer := CN_RAM_WIDTH;
			width_address	: integer := RAM_CN_ADDRESS_WIDTH
		);
		PORT (
			clock			: IN std_logic;
			enable			: IN std_logic;
			reset			: IN std_logic;
			rw_enable		: IN std_logic;
			address			: IN std_logic_vector(width_address - 1 downto 0);
			input     		: IN std_logic_vector(width_data - 1 downto 0);
			output    		: OUT std_logic_vector(width_data - 1 downto 0)
		);
	END COMPONENT;
	
	COMPONENT ShiftRegister
		GENERIC (
			size 			: integer := 1;
			width 			: integer := 32
		);
		PORT (
			clock			: IN std_logic;
			enable			: IN std_logic;
			reset			: IN std_logic;
			input     		: IN std_logic_vector(width-1 downto 0);
			output    		: OUT std_logic_vector(width-1 downto 0)
		);
	END COMPONENT;
	
	COMPONENT VariableShiftRegisterCN
	GENERIC (
		size 				: integer := MAX_DEGREE
	);
    PORT (
		clock				: IN std_logic;
		enable				: IN std_logic;
		reset				: IN std_logic;
		active_size 		: IN std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		read_enable 		: IN std_logic;
        input     			: IN std_logic;
        output    			: OUT std_logic;
		content				: OUT std_logic_vector(size - 1 downto 0)
    );
	END COMPONENT;
	
	signal memory_opt_input 			: std_logic_vector(0 to CN_RAM_WIDTH - 1);
	signal memory_opt_output 			: std_logic_vector(CN_RAM_WIDTH - 1 downto 0);
	signal oldExtrinsic			 		: std_logic_vector(WIDTH_SD - 1 downto 0);
	signal input				 		: std_logic_vector(WIDTH_SD - 1 downto 0);
	
	signal signSD 						: std_logic;
	signal signSD_total 				: std_logic;
	signal sign_of_input_message 		: std_logic;
	signal sign_total 					: std_logic;
	signal sign_total_calculated 		: std_logic;
	signal sign_of_output 				: std_logic;
	signal magnitude_of_input_message	: std_logic_vector(WIDTH_MAGNITUDE - 1 downto 0);
	signal min1							: std_logic_vector(WIDTH_MAGNITUDE - 1 downto 0);
	signal min2							: std_logic_vector(WIDTH_MAGNITUDE - 1 downto 0);
	signal min1_calculated				: std_logic_vector(WIDTH_MAGNITUDE - 1 downto 0);
	signal min2_calculated				: std_logic_vector(WIDTH_MAGNITUDE - 1 downto 0);
	signal index_min					: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal index_min_calculated			: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal indexIn_delayed				: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal sizeOfShiftRegister			: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
	signal signsToSave					: std_logic_vector(MAX_DEGREE - 1 downto 0);
	signal oldExtrinsic_index			: std_logic_vector(DEGREE_WIDTH - 1 downto 0);
		
BEGIN
		Extrinsic_Memory : MemorySinglePort
	PORT MAP (
		clock		=> clock,
		enable		=> in_enableMEM,
		reset		=> reset,
		rw_enable	=> in_readMEM,
		address		=> in_addressMEM,
		input     	=> memory_opt_input,
		output    	=> memory_opt_output
	);
	
		delay_VariableDelay : ShiftRegister
	GENERIC MAP(
		size 		=> 3,
		width 		=> DEGREE_WIDTH
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
        input		=> in_variableDelay,
        output		=> sizeOfShiftRegister
    );
	
		delay_indexIn : ShiftRegister
	GENERIC MAP(
		size 		=> 3,
		width 		=> DEGREE_WIDTH
	)
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
        input		=> in_indexIn,
        output		=> indexIn_delayed
    );
	
		delay_Signs : VariableShiftRegisterCN
    PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
        input		=> sign_of_input_message,
        output		=> sign_of_output,
		active_size => sizeOfShiftRegister,
		read_enable => in_lastValue,
		content		=> signsToSave
    );
	
	OldExtrinsicCalculation: process (clock)
	variable magnitude : std_logic_vector(WIDTH_MAGNITUDE - 1 downto 0);
	variable extrinsic : std_logic_vector(WIDTH_SD - 1 downto 0);
	BEGIN
		if clock'event and clock = '1' then
			oldExtrinsic_index <= std_logic_vector(unsigned(in_variableDelay) - unsigned(in_indexIn));
			if memory_opt_output(CN_RAM_WIDTH - 2 * WIDTH_MAGNITUDE - 1 downto CN_RAM_WIDTH - 2 * WIDTH_MAGNITUDE - DEGREE_WIDTH) = oldExtrinsic_index then
				magnitude := memory_opt_output(CN_RAM_WIDTH - WIDTH_MAGNITUDE - 1 downto CN_RAM_WIDTH - 2 * WIDTH_MAGNITUDE);
			else
				magnitude := memory_opt_output(CN_RAM_WIDTH - 1 downto CN_RAM_WIDTH - WIDTH_MAGNITUDE);
			end if;
			-- if memory_opt_output(0) = memory_opt_output(to_integer(unsigned('0' & oldExtrinsic_index) + 1)) then
				-- oldExtrinsic <= ('0' & magnitude);
			-- else
				-- oldExtrinsic <= std_logic_vector(-SD_TYPE(('0' & magnitude)));
			-- end if;
			if memory_opt_output(0) = memory_opt_output(to_integer(unsigned('0' & oldExtrinsic_index) + 1)) then
				extrinsic := std_logic_vector(-SD_TYPE(('0' & magnitude)));
			else
				extrinsic := ('0' & magnitude);
			end if;
			if in_firstIteration = '1' then
				oldExtrinsic <= (OTHERS => '0');
			else
				oldExtrinsic <= extrinsic;
			end if;
		end if;
	end process;
	
	OldSignedMagnitudeCalculation: process (clock)
	variable RnmMessage : std_logic_vector(WIDTH_SD - 1 downto 0);
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				magnitude_of_input_message <= (OTHERS => '0');
				input <= (OTHERS => '0');
			else
				input <= in_input;
				RnmMessage := std_logic_vector(SD_TYPE(input) + SD_TYPE(oldExtrinsic));
				IF RnmMessage(WIDTH_SD - 1) = '0' then
					sign_of_input_message <= '0';
					magnitude_of_input_message <= RnmMessage(WIDTH_MAGNITUDE - 1 downto 0);
				else
					sign_of_input_message <= '1';
					RnmMessage := std_logic_vector(-SD_TYPE(RnmMessage));
					magnitude_of_input_message <= RnmMessage(WIDTH_MAGNITUDE - 1 downto 0);
				end if;
			end if;
		end if;
	end process;
	
	GlobalSignCalculation: process (clock)
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				sign_total <= '0';
			end if;
			if in_firstValue = '1' then
				sign_total <= sign_of_input_message;
			elsif in_firstValue = '0' then
				sign_total <= sign_total xor sign_of_input_message;
			end if;
		end if;
	end process;
	
	ValidityCalculation: process (clock)
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				signSD 			<= '0';
				signSD_total 	<= '0';
			end if;
			signSD <= input(WIDTH_SD - 1);
			if in_firstValue = '1' then
				signSD_total <= signSD;
			elsif in_firstValue = '0' then
				signSD_total <= signSD_total xor signSD;
			end if;
		end if;
	end process;
	
	ExtensionRegisters: process (clock)
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				min1_calculated 		<= (OTHERS => '0');
				min2_calculated 		<= (OTHERS => '0');
				index_min_calculated	<= (OTHERS => '0');
				out_valid				<= '0';
				sign_total_calculated	<= '0';
			elsif in_lastValue = '1' then
				out_valid <= not signSD_total;
				sign_total_calculated <= sign_total;
				min1_calculated <= min1;
				min2_calculated <= min2;
				index_min_calculated <= index_min;
			end if;
		end if;
	end process;
	memory_opt_input <= (min1_calculated & min2_calculated & index_min_calculated & signsToSave & sign_total_calculated);
	
	NewExtrinsicCalculation: process (clock)
	variable magnitude : std_logic_vector(WIDTH_MAGNITUDE - 1 downto 0);
	BEGIN
		if clock'event and clock = '1' then
			if index_min_calculated = in_indexOut then
				magnitude := min2_calculated;
			else
				magnitude := min1_calculated;
			end if;
			if (sign_total_calculated xor sign_of_output) = '1' then
				out_output <= std_logic_vector(-SD_TYPE(('0' & magnitude)));
			else
				out_output <= ('0' & magnitude);
			end if;
		end if;
	end process;
	
		MinimaCalculation: process (clock)
	--variable en1 : boolean;
	--variable en2 : boolean;
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				min1 <= (OTHERS => '1');
				min2 <= (OTHERS => '1');
			else
				--en1 := ;
				--en2 := magnitude_of_input_message < min2;
				if magnitude_of_input_message < min1 then
					min2 <= min1;
				elsif magnitude_of_input_message < min2 then
					min2 <= magnitude_of_input_message;
				end if;
				if in_firstValue = '1' then
					min2 <= (OTHERS => '1');
				end if;
				if in_firstValue = '1' or magnitude_of_input_message < min1 then
					min1 <= magnitude_of_input_message;
					index_min <= indexIn_delayed;
				end if;
			end if;
		end if;
	end process;
	
	-- MinimaCalculation: process (clock)
	-- variable en1 : boolean;
	-- variable en2 : boolean;
	-- BEGIN
		-- if clock'event and clock = '1' then
			-- if reset = '1' then
				-- min1 <= (OTHERS => '1');
				-- min2 <= (OTHERS => '1');
			-- else
				-- en1 := magnitude_of_input_message < min1;
				-- en2 := magnitude_of_input_message < min2;
				-- if en1 = true then
					-- min2 <= min1;
				-- elsif en2 = true then
					-- min2 <= magnitude_of_input_message;
				-- end if;
				-- if in_firstValue = '1' then
					-- min2 <= (OTHERS => '1');
				-- end if;
				-- if in_firstValue = '1' or en1 = true then
					-- min1 <= magnitude_of_input_message;
					-- index_min <= indexIn_delayed;
				-- end if;
			-- end if;
		-- end if;
	-- end process;
	
		-- en1 <= magnitude_of_input_message < min1;
	-- en2 <= magnitude_of_input_message < min2;
	-- MinimaCalculation1: process (clock)
	-- BEGIN
		-- if clock'event and clock = '1' then
			-- if reset = '1' then
				-- min1		<= (OTHERS => '0');
				-- index_min	<= (OTHERS => '0');
			-- else 
				-- if in_firstValue = '1' or en1 = true then
					-- min1 <= magnitude_of_input_message;
					-- index_min <= indexIn_delayed;
				-- end if;
			-- end if;
		-- end if;
	-- end process;
	
	-- MinimaCalculation2: process (clock)
	-- BEGIN
		-- if clock'event and clock = '1' then
			-- if reset = '1' or in_firstValue = '1' then
				-- min2 <= (OTHERS => '1');
			-- elsif en1 = true then
				-- min2 <= min1;
			-- elsif en2 = true then
				-- min2 <= magnitude_of_input_message;
			-- end if;
		-- end if;
	-- end process;
END beh;