-- Filename    	: InitAndOutput.vhd
-- Author:	   	: Andreas Torno
-- Description 	: This component generates control signals and reads addresses from control memory
--				: for initialization and output states 

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;
library work;
USE work.widths.all;
USE work.delays.all;

ENTITY InitAndOutput IS
    PORT (
		-- Global control ports
		clock						: IN  std_logic;
		enable						: IN  std_logic;
		reset						: IN  std_logic;
		-- Specific control ports
		in_initState				: IN  std_logic;
		in_outputState				: IN  std_logic;
		out_loadAndOutputDone		: OUT std_logic;
		out_init					: OUT std_logic_vector(PARALLELISM - 1 downto 0);
		out_output					: OUT std_logic_vector(PARALLELISM - 1 downto 0);
		out_address					: OUT std_logic_vector(ROM_SN_WIDTH - 3 downto 0)
    );
END InitAndOutput;

ARCHITECTURE RTL OF InitAndOutput IS
	COMPONENT ROM IS
	GENERIC (
		size 				: integer := SN_NUMBER;
		width_data 			: integer := (ROM_SN_WIDTH - 2) + PARALLELISM_WIDTH;
		width_address		: integer := ROM_INITILIZATION_ADDRESS_WIDTH;
		filename			: string  := "RomInit"
	);
    PORT (
		clock		: IN std_logic;
		enable		: IN std_logic;
		reset		: IN std_logic;
		r_enable	: IN std_logic;
		address		: IN std_logic_vector(width_address - 1 downto 0);
        output    	: OUT std_logic_vector(width_data - 1 downto 0)
    );
	END COMPONENT;
	
	COMPONENT ShiftRegister
	GENERIC (
		size 		: integer := DURATION_ROM_ACCESS - 1;
		width 		: integer := 1
	);
    PORT (
		-- Control ports
		clock		: IN std_logic;
		enable		: IN std_logic;
		reset		: IN std_logic;
		
		-- Data ports
        input     	: IN std_logic_vector(width-1 downto 0);
        output    	: OUT std_logic_vector(width-1 downto 0)
    );
	END COMPONENT;
	
	signal controlMEM_address		: std_logic_vector(ROM_INITILIZATION_ADDRESS_WIDTH - 1 downto 0);
	signal controlMEM_output		: std_logic_vector((ROM_SN_WIDTH - 2) + PARALLELISM_WIDTH - 1 downto 0);
	signal controlMEM_enable		: std_logic;
	signal counter					: unsigned(ROM_INITILIZATION_ADDRESS_WIDTH - 1 downto 0);
	signal initState				: std_logic;
	signal outputState				: std_logic;
	signal in_initState_reg			: std_logic;
	signal in_outputState_reg		: std_logic;
BEGIN

	Initialization_Memory : ROM
	PORT MAP (
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
		r_enable	=> controlMEM_enable,
		address		=> controlMEM_address,
		output    	=> controlMEM_output
	);
	controlMEM_enable		<= in_initState or in_outputState;
	controlMEM_address		<= std_logic_vector(counter);
	
	count : process(clock)
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' or counter = (SN_NUMBER - 1) then
				counter					<= (OTHERS => '0');
				out_loadAndOutputDone 	<= '0';
			else
				if controlMEM_enable = '1' then
					out_loadAndOutputDone <= '0';
					if counter = (SN_NUMBER - 3) then
						out_loadAndOutputDone <= '1';
					end if;
					counter <= counter + 1;
				end if;
			end if;
		end if;
	END process;
	
	delay_initState : ShiftRegister 
	PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
		input(0)	=> in_initState_reg,
		output(0)	=> initState
	);
	
	delay_outputState : ShiftRegister 
	PORT MAP(
		clock		=> clock,
		enable		=> enable,
		reset		=> reset,
		input(0)	=> in_outputState_reg,
		output(0)	=> outputState
	);
	
	OutputROM : process(clock)
	variable bank : unsigned(PARALLELISM_WIDTH - 1 downto 0);
	BEGIN
		if clock'event and clock = '1' then
			if reset = '1' then
				out_init		<= (OTHERS => '0');
				out_output		<= (OTHERS => '0');
				out_address		<= (OTHERS => '0');
				in_outputState_reg <= '0';
				in_initState_reg   <= '0';
			else
				in_initState_reg	<= in_initState;
				in_outputState_reg	<= in_outputState;
				-- preload
				out_init		<= (OTHERS => '0');
				out_output		<= (OTHERS => '0');
				-- output of memory
				out_address		<= controlMEM_output(ROM_SN_WIDTH - 3 downto 0);
				bank			:= unsigned(controlMEM_output((ROM_SN_WIDTH - 2) + PARALLELISM_WIDTH - 1 downto ROM_SN_WIDTH - 2));
				if initState = '1' then
					out_init(to_integer(bank))		<= '1';
				end if;
				if outputState = '1' then
					out_output(to_integer(bank))	<= '1';
				end if;
			end if;
		end if;
	END process;
	
END RTL;